"""Type price.quantity.unitless, version 000"""
import json
from typing import Any
from typing import Dict
from typing import Literal

from gridworks.errors import SchemaError
from pydantic import BaseModel
from pydantic import Field
from pydantic import validator


class PriceQuantityUnitless(BaseModel):
    """ """

    PriceTimes1000: int = Field(
        title="PriceTimes1000",
    )
    QuantityTimes1000: int = Field(
        title="QuantityTimes1000",
    )
    TypeName: Literal["price.quantity.unitless"] = "price.quantity.unitless"
    Version: str = "000"

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))  # noqa


class PriceQuantityUnitless_Maker:
    type_name = "price.quantity.unitless"
    version = "000"

    def __init__(self, price_times1000: int, quantity_times1000: int):
        self.tuple = PriceQuantityUnitless(
            PriceTimes1000=price_times1000,
            QuantityTimes1000=quantity_times1000,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: PriceQuantityUnitless) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> PriceQuantityUnitless:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> PriceQuantityUnitless:
        d2 = dict(d)
        if "PriceTimes1000" not in d2.keys():
            raise SchemaError(f"dict {d2} missing PriceTimes1000")
        if "QuantityTimes1000" not in d2.keys():
            raise SchemaError(f"dict {d2} missing QuantityTimes1000")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return PriceQuantityUnitless(
            PriceTimes1000=d2["PriceTimes1000"],
            QuantityTimes1000=d2["QuantityTimes1000"],
            TypeName=d2["TypeName"],
            Version="000",
        )



"""Type price.quantity, version 000"""
import json
from enum import auto
from typing import Any
from typing import Dict
from typing import List
from typing import Literal

from fastapi_utils.enums import StrEnum
from gridworks.errors import SchemaError
from gridworks.message import as_enum
from pydantic import BaseModel
from pydantic import Field
from pydantic import validator

from gwatn.enums import MarketPriceUnit
from gwatn.enums import MarketQuantityUnit


class MarketPriceUnit000SchemaEnum:
    enum_name: str = "market.price.unit.000"
    symbols: List[str] = [
        "00000000",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class MarketPriceUnit000(StrEnum):
    USDPerMWh = auto()

    @classmethod
    def default(cls) -> "MarketPriceUnit000":
        return cls.USDPerMWh

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class MarketPriceUnitMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> MarketPriceUnit:
        if not MarketPriceUnit000SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to MarketPriceUnit000 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, MarketPriceUnit, MarketPriceUnit.default())

    @classmethod
    def local_to_type(cls, market_price_unit: MarketPriceUnit) -> str:
        if not isinstance(market_price_unit, MarketPriceUnit):
            raise SchemaError(f"{market_price_unit} must be of type {MarketPriceUnit}")
        versioned_enum = as_enum(
            market_price_unit, MarketPriceUnit000, MarketPriceUnit000.default()
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, MarketPriceUnit000] = {
        "00000000": MarketPriceUnit000.USDPerMWh,
    }

    versioned_enum_to_type_dict: Dict[MarketPriceUnit000, str] = {
        MarketPriceUnit000.USDPerMWh: "00000000",
    }


class MarketQuantityUnit000SchemaEnum:
    enum_name: str = "market.quantity.unit.000"
    symbols: List[str] = [
        "00000000",
        "c272f3b3",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class MarketQuantityUnit000(StrEnum):
    AvgMW = auto()
    AvgkW = auto()

    @classmethod
    def default(cls) -> "MarketQuantityUnit000":
        return cls.AvgMW

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class MarketQuantityUnitMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> MarketQuantityUnit:
        if not MarketQuantityUnit000SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to MarketQuantityUnit000 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, MarketQuantityUnit, MarketQuantityUnit.default())

    @classmethod
    def local_to_type(cls, market_quantity_unit: MarketQuantityUnit) -> str:
        if not isinstance(market_quantity_unit, MarketQuantityUnit):
            raise SchemaError(
                f"{market_quantity_unit} must be of type {MarketQuantityUnit}"
            )
        versioned_enum = as_enum(
            market_quantity_unit, MarketQuantityUnit000, MarketQuantityUnit000.default()
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, MarketQuantityUnit000] = {
        "00000000": MarketQuantityUnit000.AvgMW,
        "c272f3b3": MarketQuantityUnit000.AvgkW,
    }

    versioned_enum_to_type_dict: Dict[MarketQuantityUnit000, str] = {
        MarketQuantityUnit000.AvgMW: "00000000",
        MarketQuantityUnit000.AvgkW: "c272f3b3",
    }


class PriceQuantity(BaseModel):
    """ """

    PriceTimes1000: int = Field(
        title="PriceTimes1000",
    )
    QuantityTimes1000: int = Field(
        title="QuantityTimes1000",
    )
    PriceUnit: MarketPriceUnit = Field(
        title="PriceUnit",
    )
    QuantityUnit: MarketQuantityUnit = Field(
        title="QuantityUnit",
    )
    InjectionIsPositive: bool = Field(
        title="InjectionIsPositive",
    )
    TypeName: Literal["price.quantity"] = "price.quantity"
    Version: str = "000"

    @validator("PriceUnit")
    def _check_price_unit(cls, v: MarketPriceUnit) -> MarketPriceUnit:
        return as_enum(v, MarketPriceUnit, MarketPriceUnit.USDPerMWh)

    @validator("QuantityUnit")
    def _check_quantity_unit(cls, v: MarketQuantityUnit) -> MarketQuantityUnit:
        return as_enum(v, MarketQuantityUnit, MarketQuantityUnit.AvgMW)

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        del d["PriceUnit"]
        PriceUnit = as_enum(self.PriceUnit, MarketPriceUnit, MarketPriceUnit.default())
        d["PriceUnitGtEnumSymbol"] = MarketPriceUnitMap.local_to_type(PriceUnit)
        del d["QuantityUnit"]
        QuantityUnit = as_enum(
            self.QuantityUnit, MarketQuantityUnit, MarketQuantityUnit.default()
        )
        d["QuantityUnitGtEnumSymbol"] = MarketQuantityUnitMap.local_to_type(
            QuantityUnit
        )
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))  # noqa


class PriceQuantity_Maker:
    type_name = "price.quantity"
    version = "000"

    def __init__(
        self,
        price_times1000: int,
        quantity_times1000: int,
        price_unit: MarketPriceUnit,
        quantity_unit: MarketQuantityUnit,
        injection_is_positive: bool,
    ):
        self.tuple = PriceQuantity(
            PriceTimes1000=price_times1000,
            QuantityTimes1000=quantity_times1000,
            PriceUnit=price_unit,
            QuantityUnit=quantity_unit,
            InjectionIsPositive=injection_is_positive,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: PriceQuantity) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> PriceQuantity:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> PriceQuantity:
        d2 = dict(d)
        if "PriceTimes1000" not in d2.keys():
            raise SchemaError(f"dict {d2} missing PriceTimes1000")
        if "QuantityTimes1000" not in d2.keys():
            raise SchemaError(f"dict {d2} missing QuantityTimes1000")
        if "PriceUnitGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing PriceUnitGtEnumSymbol")
        if d2["PriceUnitGtEnumSymbol"] in MarketPriceUnit000SchemaEnum.symbols:
            d2["PriceUnit"] = MarketPriceUnitMap.type_to_local(
                d2["PriceUnitGtEnumSymbol"]
            )
        else:
            d2["PriceUnit"] = MarketPriceUnit.default()
        if "QuantityUnitGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing QuantityUnitGtEnumSymbol")
        if d2["QuantityUnitGtEnumSymbol"] in MarketQuantityUnit000SchemaEnum.symbols:
            d2["QuantityUnit"] = MarketQuantityUnitMap.type_to_local(
                d2["QuantityUnitGtEnumSymbol"]
            )
        else:
            d2["QuantityUnit"] = MarketQuantityUnit.default()
        if "InjectionIsPositive" not in d2.keys():
            raise SchemaError(f"dict {d2} missing InjectionIsPositive")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return PriceQuantity(
            PriceTimes1000=d2["PriceTimes1000"],
            QuantityTimes1000=d2["QuantityTimes1000"],
            PriceUnit=d2["PriceUnit"],
            QuantityUnit=d2["QuantityUnit"],
            InjectionIsPositive=d2["InjectionIsPositive"],
            TypeName=d2["TypeName"],
            Version="000",
        )



import logging
from typing import List

import dotenv
import gridworks.algo_utils as algo_utils
import gridworks.api_utils as api_utils
from algosdk import encoding
from algosdk.future import transaction
from algosdk.v2client.algod import AlgodClient
from gridworks.algo_utils import BasicAccount
from gridworks.utils import RestfulResponse

import gwatn.config as config
from gwatn.enums import AlgoCertType
from gwatn.types import InitialTadeedAlgoOptin
from gwatn.types import NewTadeedAlgoOptin
from gwatn.types import NewTadeedSend_Maker
from gwatn.types import OldTadeedAlgoReturn
from gwatn.types import SlaEnter


LOGGER = logging.getLogger(__name__)
DUMMY_ACCT_ADDR = "NZXUSTZACPVJBHRSSJ5KE3JUPCITK5P2O4FE67NYPXRDVCJA6ZX4AL62EA"


class PythonTaDaemon:
    def __init__(
        self,
        settings: config.TaDaemonSettings = config.TaDaemonSettings(
            _env_file=dotenv.find_dotenv()
        ),
    ):
        self.settings = settings
        self.client: AlgodClient = AlgodClient(
            settings.algo_api_secrets.algod_token.get_secret_value(),
            settings.public.algod_address,
        )
        self.acct: BasicAccount = BasicAccount(
            private_key=self.settings.sk.get_secret_value()
        )
        self.trading_rights_addr: str = self.acct.addr
        LOGGER.info("TaOwner Smart Daemon Initialized")

    ##########################
    # Messages Received
    ##########################

    def ta_deed_icon_file(self) -> str:
        if self.has_deed():
            return "violet_icon.png"
        else:
            return "not_a_plant.png"

    def initial_tadeed_algo_optin_received(
        self, payload: InitialTadeedAlgoOptin
    ) -> RestfulResponse:
        if self.has_deed():
            r = RestfulResponse(
                Note=f"Ignoring InitialTadeed Optin. Deeds: {self.ta_deed_alias_list}"
            )
            return r
        ta_deed_id = api_utils.get_tadeed_id(
            terminal_asset_alias=payload.TerminalAssetAlias,
            validator_addr=payload.ValidatorAddr,
        )
        if ta_deed_id is None:
            note = f"called when validator {payload.ValidatorAddr[-6:]} did NOT have TADEED for {payload.TerminalAssetAlias}!"
            LOGGER.info(note)
            r = RestfulResponse(
                Note=note,
                HttpStatusCode=422,
            )
            return r
        if ta_deed_id.Type == AlgoCertType.SmartSig:
            raise NotImplementedError(f"Does not yet handle SmartSig TaDeeds")
        # Call a TaDaemonSmartContract method of the same
        # name (InitialTadeedAlgoOptin, in some format)
        # args should include :
        #   - the asset-index
        #   - the TaOwnerAddress
        #   - the TaValidatorAddress
        #
        # the TaDaemonSmartContract should check that:
        #
        #  - That it does not own any assets yet (this is an initial transfer)
        #  - the TaOwnerAddress is its TaOwnerAddress (which it needs to be initialized with)
        #  - the TaValidatorAddress is its TaValidatorAddress (which it needs to be initialized with)
        txn = transaction.AssetOptInTxn(
            sender=self.acct.addr,
            index=ta_deed_id.Idx,
            sp=self.client.suggested_params(),
        )
        signed_txn = txn.sign(self.acct.sk)
        try:
            self.client.send_transaction(signed_txn)
        except:
            return RestfulResponse(
                Note=f"Failure sending transaction to opt into TaDeed {ta_deed_id.Idx}",
                HttpStatusCode=422,
            )
        algo_utils.wait_for_transaction(self.client, signed_txn.get_txid())
        ta_trading_rights_id = api_utils.get_tatrading_rights_id(
            terminal_asset_alias=payload.TerminalAssetAlias
        )
        if ta_trading_rights_id.Type == AlgoCertType.SmartSig:
            raise NotImplementedError(f"Does not yet handle SmartSig TaTradingRights")
        txn = transaction.AssetOptInTxn(
            sender=self.acct.addr,
            index=ta_trading_rights_id.Idx,
            sp=self.client.suggested_params(),
        )
        signed_txn = txn.sign(self.acct.sk)
        try:
            self.client.send_transaction(signed_txn)
        except:
            return RestfulResponse(
                Note=f"Failure sending transaction to opt into TaTradingRights {ta_trading_rights_id.Idx}",
                HttpStatusCode=422,
            )

        note = (
            f"TaDaemon successfully opted in to Initial TaDeed {ta_deed_id.Idx}"
            f" and TaTradingRights {ta_trading_rights_id.Idx}"
        )
        LOGGER.info(note)
        r = RestfulResponse(Note=note)
        return r

    def new_tadeed_algo_optin_received(
        self, payload: NewTadeedAlgoOptin
    ) -> RestfulResponse:
        """Opts in to a new (i.e. updated) TaDeed

        Args:
            payload (NewTadeedAlgoOptin): NewTadeedAlgoOptin

        Returns:
            RestfulResponse:  HttpStatusCode 422 if there is a semantic
            issue (e.g. failure sending transaction on blockchain)

            Otherwise, Payload is NewTadeedSend
        """
        # Call a TaDaemonSmartContract method of the same
        # name (InitialTadeedAlgoOptin, in some format)
        # args should include :
        #   - the asset-index
        #   - the AssetCreatorAddr
        #
        #
        # the TaDaemonSmartContract should check that:
        #
        #  - That it does not own any assets yet (this is an initial transfer)
        #  - the AssetCreatorAddr is "RNMHG32VTIHTC7W3LZOEPTDGREL5IQGK46HKD3KBLZHYQUCAKLMT4G5ALI"
        #        (the GNodeFactory)
        txn = transaction.AssetOptInTxn(
            sender=self.acct.addr,
            index=payload.NewTaDeedIdx,
            sp=self.client.suggested_params(),
        )

        signed_txn = txn.sign(self.acct.sk)
        try:
            self.client.send_transaction(signed_txn)
        except:
            note = "Failure sending transaction on Algo blockchain"
            r = RestfulResponse(
                Note=note,
                HttpStatusCode=422,
            )
            return r
        algo_utils.wait_for_transaction(self.client, signed_txn.get_txid())

        return_payload = NewTadeedSend_Maker(
            new_ta_deed_idx=payload.NewTaDeedIdx,
            old_ta_deed_idx=payload.OldTaDeedIdx,
            ta_daemon_addr=self.acct.addr,
            validator_addr=payload.ValidatorAddr,
            signed_tadeed_optin_txn=encoding.msgpack_encode(signed_txn),
        ).tuple
        note = f"Opted in to TaDeed {payload.NewTaDeedIdx}, ready for transfer"
        r = RestfulResponse(
            Note=note,
            PayloadTypeName=NewTadeedSend_Maker.type_name,
            PayloadAsDict=return_payload.as_dict(),
        )
        LOGGER.info(f"Opted in to TaDeed {payload.NewTaDeedIdx}, ready for transfer")
        return r

    def old_tadeed_algo_return_received(
        self, payload: OldTadeedAlgoReturn
    ) -> RestfulResponse:
        """
         - Transfer the  old deed back to the GNodeFactory admin acct.

        Args:
            payload: OldTadeedAlgoReturn
        """

        # Call a TaDaemonSmartContract method of the same
        # name (OldTadeedAlgoReturn, in some format)
        # args should include :
        #   - the receiver
        #   - the amount
        #    - the asset_index
        # the TaDaemonSmartContract should check
        # that:
        #  - The receiver is "RNMHG32VTIHTC7W3LZOEPTDGREL5IQGK46HKD3KBLZHYQUCAKLMT4G5ALI"
        #  (The GNodeFactory admin address, which it should be initialized with)
        #  - The amount is 1
        #
        txn = transaction.AssetTransferTxn(
            sender=self.acct.addr,
            receiver=config.Public().gnf_admin_addr,
            amt=1,
            index=payload.OldTaDeedIdx,
            sp=self.client.suggested_params(),
        )
        signed_txn = txn.sign(self.acct.sk)
        try:
            self.client.send_transaction(signed_txn)
        except:
            note = "Failure sending transaction on Algo blockchain"
            r = RestfulResponse(
                Note=note,
                HttpStatusCode=422,
            )
            return r

        algo_utils.wait_for_transaction(self.client, signed_txn.get_txid())
        note = f"TaDaemon transferred old TaDeed {payload.OldTaDeedIdx} to GNodeFactoryAdmin"
        LOGGER.info(note)
        r = RestfulResponse(Note=note)
        return r

    def ta_deed_alias_list(self) -> List[str]:
        if self.acct is None:
            return []
        try:
            assets = self.client.account_info(self.acct.addr)["assets"]
        except:
            return []
        owned_assets = list(filter(lambda x: x["amount"] == 1, assets))
        owned_asset_idx_list = list(map(lambda x: x["asset-id"], owned_assets))
        deed_alias_list = []
        for asset_idx in owned_asset_idx_list:
            alias = api_utils.alias_from_deed_idx(asset_idx)
            if alias:
                deed_alias_list.append(alias)
        return deed_alias_list

    def has_deed(self) -> bool:
        if len(self.ta_deed_alias_list()) > 0:
            return True
        return False

    def sla_enter_received(self, payload: SlaEnter) -> RestfulResponse:
        ta_alias = payload.TerminalAssetAlias
        ta_trading_rights_id = api_utils.get_tatrading_rights_id(
            terminal_asset_alias=ta_alias
        )
        if ta_trading_rights_id.Type == AlgoCertType.SmartSig:
            raise NotImplementedError("Does not handle SmartSig TaTradingRights yet")
        alias_list: List[str] = [
            "d1.isone.ver.keene.holly.ta",
            "d1.isone.ver.keene.juniper.ta",
            "d1.isone.ver.keene.kale.ta",
            "d1.isone.ver.keene.lettuce.ta",
        ]

        # These are the secret keys for the 4 demo AtomicTNodes.
        sk_list = [
            "K6iB3AHmzSQ8wDE91QdUfaheDMEtf2WJUMYeeRptKxHiTxG3HC+iKpngXmi82y2r9uVPYwTI5aGiMhdXmPRxcQ==",
            "QfKe/7kzD71nhYGfITlSV/DFYGvC4sc5IEa8ieGsgirC9sBaSJT0O1+mdPOK3/wzZAqy/dRVIg58Uh3ucSIUSw==",
            "UHDv5NTx3pz26XZwpbjwKxmdnYzksEuOmbTbvzgkQbYVnLuK+VLwt0QwgJHAUODoluXS8R5InKOS2X1qNqgBeA==",
            "pMpo89JUKfRE+IvXXW/dsAkns0FXpxagXtQf4m6sXIeO3qH3lHAdbzJvd8gs+xfgnQs3oUs47KD4sWtTARNndw==",
        ]
        if ta_alias not in alias_list:
            return RestfulResponse(
                Note="Did not transfer TaTradingRights, not one of the first 4 atns"
            )
        required_algos = 25

        i = alias_list.index(ta_alias)
        sk = sk_list[i]
        atn_acct = BasicAccount(sk)

        txn = transaction.PaymentTxn(
            sender=self.acct.addr,
            receiver=atn_acct.addr,
            amt=required_algos * 10**6,
            sp=self.client.suggested_params(),
        )
        signed_txn = txn.sign(self.acct.sk)
        try:
            self.client.send_transaction(signed_txn)
        except:
            return RestfulResponse(
                Note=f"Failure Funding {ta_alias} atn",
                HttpStatusCode=422,
            )

        # Hack opt the Atn into the asset using the ATNS PRIVATE KEY.
        # Replace when real SLA exists
        txn = transaction.AssetOptInTxn(
            sender=atn_acct.addr,
            index=ta_trading_rights_id.Idx,
            sp=self.client.suggested_params(),
        )
        signed_txn = txn.sign(atn_acct.sk)
        try:
            self.client.send_transaction(signed_txn)
        except:
            return RestfulResponse(
                Note=f"Atn for {ta_alias} failed to opt into trading rights {ta_trading_rights_id.Idx}",
                HttpStatusCode=422,
            )

        txn = transaction.AssetTransferTxn(
            sender=self.acct.addr,
            receiver=atn_acct.addr,
            amt=1,
            index=ta_trading_rights_id.Idx,
            sp=self.client.suggested_params(),
        )
        signed_txn = txn.sign(self.acct.sk)
        try:
            self.client.send_transaction(signed_txn)
        except:
            note = (
                f"Failure sending AssetTransfer for {ta_alias} trading rights"
                f" {ta_trading_rights_id.Idx}"
            )
            r = RestfulResponse(Note=note, HttpStatusCode=422)
            return r
        self.trading_rights_addr = atn_acct.addr
        return RestfulResponse(
            Note=f"Successfully entered Service Level Agreement for {ta_alias}"
        )



from pydantic import BaseModel

from gwatn import property_format
from gwatn.data_classes import MarketType
from gwatn.enums import MarketTypeName
from gwatn.types import MarketSlot
from gwatn.types import MarketTypeGt_Maker


DUMMY_ALGO_TXN = "gqNzaWfEQNPXbrAiWd+cNgsIaM3N0PSu3repauvmjuHmoKjh6sd3L5U4/YpovcXN7/ATH1LgcI4cgV+SU3VQ6bsm/gfAOQyjdHhuiaNhbXTNB9CjZmVlzQPoomZ2HaNnZW6qc2FuZG5ldC12MaJnaMQgaJXPYTdWaeTNSs8FMMzPNfV7SrHXqJgFsJLxRbSPjzCibHbNBAWjcmN2xCBWkH3PValty0Rb0cyZo69Alhp4IbNKFnhXtgJ++A9EzKNzbmTEIOJPEbccL6IqmeBeaLzbLav25U9jBMjloaIyF1eY9HFxpHR5cGWjcGF5"


class DijkstraChoice(BaseModel):
    PowerImportedKw: float
    NextNodeCostDollars: float


class CostAndQuantityBought(BaseModel):
    QuantityBought: float
    Cost: float


def name_from_market_slot(slot: MarketSlot) -> str:
    return f"{slot.Type.Name.value}.{slot.MarketMakerAlias}.{slot.StartUnixS}"


def market_slot_from_name(market_slot_name: str) -> MarketSlot:
    """rt60gate30b.d1.isone.ver.keene.1577836800"""
    try:
        property_format.check_is_market_slot_name_lrd_format(market_slot_name)
    except ValueError as e:
        raise Exception(
            f"market slot alias {market_slot_name} does not have market"
            " slot alias lrd format!"
        )
    words = market_slot_name.split(".")
    market_type_name = MarketTypeName(words[0])
    market_type_dc = MarketType.by_id[market_type_name]
    market_type = MarketTypeGt_Maker.dc_to_tuple(market_type_dc)
    market_maker_alias = ".".join(words[1:-1])
    slot_start = int(words[-1])
    return MarketSlot(
        Type=market_type, MarketMakerAlias=market_maker_alias, StartUnixS=slot_start
    )



"""Gridworks AtomicTNode"""
import gridworks.errors as errors
import gridworks.property_format as property_format

import gwatn.atn_utils as atn_utils
import gwatn.config as config
import gwatn.enums as enums
from gwatn.atn_actor_base import AtnActorBase
from gwatn.dispatch_contract import DispatchContract
from gwatn.two_channel_actor_base import TwoChannelActorBase


__all__ = [
    "atn_utils",
    "config",
    "enums",
    "property_format",
    "errors",
    "AtnActorBase",
    "DispatchContract",
    "TwoChannelActorBase",
]



"""Type flo.params.report, version 000"""
import json
from typing import Any
from typing import Dict
from typing import Literal
from typing import Optional

from gridworks.errors import SchemaError
from pydantic import BaseModel
from pydantic import Field
from pydantic import validator

from gwatn.types.flo_params import FloParams
from gwatn.types.flo_params import FloParams_Maker


def check_is_reasonable_unix_time_s(v: int) -> None:
    """
    ReasonableUnixTimeS format: time in unix seconds between Jan 1 2000 and Jan 1 3000

    Raises:
        ValueError: if not ReasonableUnixTimeS format
    """
    import pendulum

    if pendulum.parse("2000-01-01T00:00:00Z").int_timestamp > v:  # type: ignore[attr-defined]
        raise ValueError(f"{v} must be after Jan 1 2000")
    if pendulum.parse("3000-01-01T00:00:00Z").int_timestamp < v:  # type: ignore[attr-defined]
        raise ValueError(f"{v} must be before Jan 1 3000")


def check_is_uuid_canonical_textual(v: str) -> None:
    """
    UuidCanonicalTextual format:  A string of hex words separated by hyphens
    of length 8-4-4-4-12.

    Raises:
        ValueError: if not UuidCanonicalTextual format
    """
    try:
        x = v.split("-")
    except AttributeError as e:
        raise ValueError(f"Failed to split on -: {e}")
    if len(x) != 5:
        raise ValueError(f"{v} split by '-' did not have 5 words")
    for hex_word in x:
        try:
            int(hex_word, 16)
        except ValueError:
            raise ValueError(f"Words of {v} are not all hex")
    if len(x[0]) != 8:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[1]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[2]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[3]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[4]) != 12:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")


def check_is_left_right_dot(v: str) -> None:
    """
    LeftRightDot format: Lowercase alphanumeric words separated by periods,
    most significant word (on the left) starting with an alphabet character.

    Raises:
        ValueError: if not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate {v} into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(f"Most significant word of {v} must start with alphabet char.")
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of {v} split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of {v} must be lowercase.")


class FloParamsReport(BaseModel):
    """Flo Params Report.

    Type used for a message provided by an AtomicTNode or SCADA actor re the flo parameters just used to run a FLO.
    [More info](https://gridworks-atn.readthedocs.io/en/latest/flo.html#flo-params).
    """

    GNodeAlias: str = Field(
        title="GNodeAlias",
    )
    GNodeInstanceId: str = Field(
        title="GNodeInstanceId",
    )
    FloParamsTypeName: str = Field(
        title="FloParamsTypeName",
    )
    FloParamsTypeVersion: str = Field(
        title="FloParamsTypeVersion",
    )
    ReportGeneratedTimeUnixS: int = Field(
        title="ReportGeneratedTimeUnixS",
    )
    IrlTimeUnixS: Optional[int] = Field(
        title="IrlTimeUnixS",
        default=None,
    )
    Params: FloParams = Field(
        title="Params",
    )
    TypeName: Literal["flo.params.report"] = "flo.params.report"
    Version: str = "000"

    @validator("GNodeAlias")
    def _check_g_node_alias(cls, v: str) -> str:
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(f"GNodeAlias failed LeftRightDot format validation: {e}")
        return v

    @validator("GNodeInstanceId")
    def _check_g_node_instance_id(cls, v: str) -> str:
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"GNodeInstanceId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @validator("FloParamsTypeName")
    def _check_flo_params_type_name(cls, v: str) -> str:
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(
                f"FloParamsTypeName failed LeftRightDot format validation: {e}"
            )
        return v

    @validator("ReportGeneratedTimeUnixS")
    def _check_report_generated_time_unix_s(cls, v: int) -> int:
        try:
            check_is_reasonable_unix_time_s(v)
        except ValueError as e:
            raise ValueError(
                f"ReportGeneratedTimeUnixS failed ReasonableUnixTimeS format validation: {e}"
            )
        return v

    @validator("IrlTimeUnixS")
    def _check_irl_time_unix_s(cls, v: Optional[int]) -> Optional[int]:
        if v is None:
            return v
        try:
            check_is_reasonable_unix_time_s(v)
        except ValueError as e:
            raise ValueError(
                f"IrlTimeUnixS failed ReasonableUnixTimeS format validation: {e}"
            )
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        if d["IrlTimeUnixS"] is None:
            del d["IrlTimeUnixS"]
        d["Params"] = self.Params.as_dict()
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))  # noqa


class FloParamsReport_Maker:
    type_name = "flo.params.report"
    version = "000"

    def __init__(
        self,
        g_node_alias: str,
        g_node_instance_id: str,
        flo_params_type_name: str,
        flo_params_type_version: str,
        report_generated_time_unix_s: int,
        irl_time_unix_s: Optional[int],
        params: FloParams,
    ):
        self.tuple = FloParamsReport(
            GNodeAlias=g_node_alias,
            GNodeInstanceId=g_node_instance_id,
            FloParamsTypeName=flo_params_type_name,
            FloParamsTypeVersion=flo_params_type_version,
            ReportGeneratedTimeUnixS=report_generated_time_unix_s,
            IrlTimeUnixS=irl_time_unix_s,
            Params=params,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: FloParamsReport) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> FloParamsReport:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> FloParamsReport:
        d2 = dict(d)
        if "GNodeAlias" not in d2.keys():
            raise SchemaError(f"dict {d2} missing GNodeAlias")
        if "GNodeInstanceId" not in d2.keys():
            raise SchemaError(f"dict {d2} missing GNodeInstanceId")
        if "FloParamsTypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing FloParamsTypeName")
        if "FloParamsTypeVersion" not in d2.keys():
            raise SchemaError(f"dict {d2} missing FloParamsTypeVersion")
        if "ReportGeneratedTimeUnixS" not in d2.keys():
            raise SchemaError(f"dict {d2} missing ReportGeneratedTimeUnixS")
        if "IrlTimeUnixS" not in d2.keys():
            d2["IrlTimeUnixS"] = None
        if "Params" not in d2.keys():
            raise SchemaError(f"dict {d2} missing Params")
        if not isinstance(d2["Params"], dict):
            raise SchemaError(f"d['Params'] {d2['Params']} must be a FloParams!")
        params = FloParams_Maker.dict_to_tuple(d2["Params"])
        d2["Params"] = params
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return FloParamsReport(
            GNodeAlias=d2["GNodeAlias"],
            GNodeInstanceId=d2["GNodeInstanceId"],
            FloParamsTypeName=d2["FloParamsTypeName"],
            FloParamsTypeVersion=d2["FloParamsTypeVersion"],
            ReportGeneratedTimeUnixS=d2["ReportGeneratedTimeUnixS"],
            IrlTimeUnixS=d2["IrlTimeUnixS"],
            Params=d2["Params"],
            TypeName=d2["TypeName"],
            Version="000",
        )



"""Type dispatch.contract.confirmed, version 000"""
import json
from typing import Any
from typing import Dict
from typing import Literal

from gridworks.errors import SchemaError
from pydantic import BaseModel
from pydantic import Field
from pydantic import root_validator
from pydantic import validator

from gwatn.types.atn_params import AtnParams
from gwatn.types.atn_params import AtnParams_Maker


def check_is_uuid_canonical_textual(v: str) -> None:
    """
    UuidCanonicalTextual format:  A string of hex words separated by hyphens
    of length 8-4-4-4-12.

    Raises:
        ValueError: if not UuidCanonicalTextual format
    """
    try:
        x = v.split("-")
    except AttributeError as e:
        raise ValueError(f"Failed to split on -: {e}")
    if len(x) != 5:
        raise ValueError(f"{v} split by '-' did not have 5 words")
    for hex_word in x:
        try:
            int(hex_word, 16)
        except ValueError:
            raise ValueError(f"Words of {v} are not all hex")
    if len(x[0]) != 8:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[1]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[2]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[3]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[4]) != 12:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")


def check_is_left_right_dot(v: str) -> None:
    """
    LeftRightDot format: Lowercase alphanumeric words separated by periods,
    most significant word (on the left) starting with an alphabet character.

    Raises:
        ValueError: if not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate {v} into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(f"Most significant word of {v} must start with alphabet char.")
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of {v} split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of {v} must be lowercase.")


def check_is_algo_msg_pack_encoded(v: str) -> None:
    """
    AlgoMSgPackEncoded format: the format of an  transaction sent to
    the Algorand blockchain.

    Raises:
        ValueError: if not AlgoMSgPackEncoded  format
    """
    import algosdk

    try:
        algosdk.encoding.future_msgpack_decode(v)
    except Exception as e:
        raise ValueError(f"Not AlgoMsgPackEncoded format: {e}")


class DispatchContractConfirmed(BaseModel):
    """Message sent from AtomicTNode back to SCADA via Rabbit .

    Paired with join.dispatch.contract. Sent from AtomicTNode back to SCADA
    once the AtomicTNode has successfully finished bootstrapping the Dispatch
    Contract and opted in. Once it has done this, the Dispatch Contract is ready
     to collect audit information about heartbeats, dispatch, energy and power.

    https://gridworks.readthedocs.io/en/latest/dispatch-contract.html
    """

    FromGNodeAlias: str = Field(
        title="FromGNodeAlias",
    )
    FromGNodeInstanceId: str = Field(
        title="FromGNodeInstanceId",
    )
    AtnParamsTypeName: str = Field(
        title="AtnParamsTypeName",
    )
    SignedProof: str = Field(
        title="SignedProof",
    )
    Params: AtnParams = Field(
        title="Params",
    )
    TypeName: Literal["dispatch.contract.confirmed"] = "dispatch.contract.confirmed"
    Version: str = "000"

    @validator("FromGNodeAlias")
    def _check_from_g_node_alias(cls, v: str) -> str:
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(
                f"FromGNodeAlias failed LeftRightDot format validation: {e}"
            )
        return v

    @validator("FromGNodeInstanceId")
    def _check_from_g_node_instance_id(cls, v: str) -> str:
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"FromGNodeInstanceId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @validator("AtnParamsTypeName")
    def _check_atn_params_type_name(cls, v: str) -> str:
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(
                f"AtnParamsTypeName failed LeftRightDot format validation: {e}"
            )
        return v

    @validator("SignedProof")
    def _check_signed_proof(cls, v: str) -> str:
        try:
            check_is_algo_msg_pack_encoded(v)
        except ValueError as e:
            raise ValueError(
                f"SignedProof failed AlgoMsgPackEncoded format validation: {e}"
            )
        return v

    @root_validator
    def check_axiom_1(cls, v: dict) -> dict:
        """
        Axiom 1: AtnParamsTypeName matches AtnParams.
        AtnParams must have
        """
        # TODO: Implement check for axiom 1"
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        d["Params"] = self.Params.as_dict()
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))  # noqa


class DispatchContractConfirmed_Maker:
    type_name = "dispatch.contract.confirmed"
    version = "000"

    def __init__(
        self,
        from_g_node_alias: str,
        from_g_node_instance_id: str,
        atn_params_type_name: str,
        signed_proof: str,
        params: AtnParams,
    ):
        self.tuple = DispatchContractConfirmed(
            FromGNodeAlias=from_g_node_alias,
            FromGNodeInstanceId=from_g_node_instance_id,
            AtnParamsTypeName=atn_params_type_name,
            SignedProof=signed_proof,
            Params=params,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: DispatchContractConfirmed) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> DispatchContractConfirmed:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> DispatchContractConfirmed:
        d2 = dict(d)
        if "FromGNodeAlias" not in d2.keys():
            raise SchemaError(f"dict {d2} missing FromGNodeAlias")
        if "FromGNodeInstanceId" not in d2.keys():
            raise SchemaError(f"dict {d2} missing FromGNodeInstanceId")
        if "AtnParamsTypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing AtnParamsTypeName")
        if "SignedProof" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SignedProof")
        if "Params" not in d2.keys():
            raise SchemaError(f"dict {d2} missing Params")
        if not isinstance(d2["Params"], dict):
            raise SchemaError(f"d['Params'] {d2['Params']} must be a AtnParams!")
        params = AtnParams_Maker.dict_to_tuple(d2["Params"])
        d2["Params"] = params
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return DispatchContractConfirmed(
            FromGNodeAlias=d2["FromGNodeAlias"],
            FromGNodeInstanceId=d2["FromGNodeInstanceId"],
            AtnParamsTypeName=d2["AtnParamsTypeName"],
            SignedProof=d2["SignedProof"],
            Params=d2["Params"],
            TypeName=d2["TypeName"],
            Version="000",
        )



import logging
import time

import dotenv
import rich
from gwproto.messages import PeerActiveEvent
from gwproto.messages import Ping as GridworksPing

import gwatn.config as config
from gwatn.atn_actor_base import AtnActorBase
from gwatn.enums import TelemetryName
from gwatn.types import GtDispatchBoolean_Maker
from gwatn.types import GtShCliAtnCmd_Maker
from gwatn.types import GtShStatus
from gwatn.types import LatestPrice
from gwatn.types import PowerWatts
from gwatn.types import SimTimestep
from gwatn.types import SnapshotSpaceheat


LOG_FORMAT = (
    "%(levelname) -10s %(sasctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)
LOGGER = logging.getLogger(__name__)


class SimpleAtnActor(AtnActorBase):
    """Simple implementation of an AtnActor, for testing purposes"""

    def __init__(
        self,
        settings: config.AtnSettings = config.AtnSettings(
            _env_file=dotenv.find_dotenv()
        ),
    ):
        super().__init__(settings=settings)
        self._power_watts: int = 0
        LOGGER.info("Simple Atn Initialized")

    def latest_price_from_market_maker(self, payload: LatestPrice) -> None:
        pass

    def new_timestep(self, payload: SimTimestep) -> None:
        """Set to work with a timestep per minute"""

        # sends a hb to Scada every minute for DispatchContract
        if self.in_dispatch_contract():
            self.hb_to_scada()

    def repeat_timestep(self, payload: SimTimestep) -> None:
        pass

    ##################
    # Sent to SCADA
    ##################

    def snap(self):
        self.send_scada_message(
            payload=GtShCliAtnCmd_Maker(
                from_g_node_alias=self.alias,
                from_g_node_id=self.g_node_instance_id,
                send_snapshot=True,
            ).tuple
        )

    def turn_on(self, relay_node_name: str) -> None:
        self.send_scada_message(
            payload=GtDispatchBoolean_Maker(
                about_node_name=relay_node_name,
                to_g_node_alias=self.scada_alias,
                from_g_node_alias=self.alias,
                from_g_node_instance_id=self.g_node_instance_id,
                relay_state=1,
                send_time_unix_ms=int(time.time() * 1000),
            ).tuple
        )

    def turn_off(self, relay_node_name: str) -> None:
        self.send_scada_message(
            payload=GtDispatchBoolean_Maker(
                about_node_name=relay_node_name,
                to_g_node_alias=self.scada_alias,
                from_g_node_alias=self.alias,
                from_g_node_instance_id=self.g_node_instance_id,
                relay_state=0,
                send_time_unix_ms=int(time.time() * 1000),
            ).tuple
        )

    ######################
    # Received from SCADA
    #####################
    def _process_gridworks_ping_from_scada(self, payload: GridworksPing) -> None:
        """Atn has received gridworks.ping message from its SCADA"""
        ...
        LOGGER.debug(f"_process_gridworks_ping_from_scada: {payload}")

    def _process_peer_active_event_from_scada(self, payload: PeerActiveEvent) -> None:
        """Atn has received gridworks.event.comm.peer.active message from its SCADA"""
        LOGGER.debug(f"_process_peer_active_event_from_scada: {payload}")

    def _process_gt_sh_status_from_scada(self, payload: GtShStatus) -> None:
        """Atn has received gt.sh.status message from its SCADA"""
        self.latest_status = payload
        LOGGER.debug(f"_process_gt_sh_status_from_scada: {payload}")

    def _process_power_watts_from_scada(self, payload: PowerWatts) -> None:
        """Atn has received power.watts message from its SCADA"""
        rich.print(f"Agg Power {payload.Watts} W")
        self._power_watts = payload.Watts

    def _process_snapshot_spaceheat_from_scada(
        self, payload: SnapshotSpaceheat
    ) -> None:
        """Atn has received gridworks.event.snapshot.spaceheat message from its SCADA"""
        s = "\n\nSnapshot received:\n"
        for i in range(len(payload.Snapshot.AboutNodeAliasList)):
            telemetry_name = payload.Snapshot.TelemetryNameList[i]
            if telemetry_name == TelemetryName.WaterTempCTimes1000:
                centigrade = payload.Snapshot.ValueList[i] / 1000
                fahrenheit = (centigrade * 9 / 5) + 32
                extra = f"{fahrenheit:5.2f} F"
            elif telemetry_name == TelemetryName.WaterTempFTimes1000:
                fahrenheit = payload.Snapshot.ValueList[i] / 1000
                extra = f"{fahrenheit:5.2f} F"
            else:
                extra = (
                    f"{payload.Snapshot.ValueList[i]} "
                    f"{payload.Snapshot.TelemetryNameList[i].value}"
                )
            s += f"  {payload.Snapshot.AboutNodeAliasList[i]}: {extra}\n"
        LOGGER.info(s)



"""Type gt.dispatch.boolean, version 110"""
import json
from typing import Any
from typing import Dict
from typing import Literal

from gridworks.errors import SchemaError
from pydantic import BaseModel
from pydantic import Field
from pydantic import validator


def check_is_uuid_canonical_textual(v: str) -> None:
    """
    UuidCanonicalTextual format:  A string of hex words separated by hyphens
    of length 8-4-4-4-12.

    Raises:
        ValueError: if not UuidCanonicalTextual format
    """
    try:
        x = v.split("-")
    except AttributeError as e:
        raise ValueError(f"Failed to split on -: {e}")
    if len(x) != 5:
        raise ValueError(f"{v} split by '-' did not have 5 words")
    for hex_word in x:
        try:
            int(hex_word, 16)
        except ValueError:
            raise ValueError(f"Words of {v} are not all hex")
    if len(x[0]) != 8:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[1]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[2]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[3]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[4]) != 12:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")


def check_is_left_right_dot(v: str) -> None:
    """
    LeftRightDot format: Lowercase alphanumeric words separated by periods,
    most significant word (on the left) starting with an alphabet character.

    Raises:
        ValueError: if not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate {v} into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(f"Most significant word of {v} must start with alphabet char.")
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of {v} split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of {v} must be lowercase.")


def check_is_reasonable_unix_time_ms(v: str) -> None:
    """
    ReasonableUnixTimeMs format: time in unix milliseconds between Jan 1 2000 and Jan 1 3000

    Raises:
        ValueError: if not ReasonableUnixTimeMs format
    """
    import pendulum

    if pendulum.parse("2000-01-01T00:00:00Z").int_timestamp * 1000 > v:  # type: ignore[attr-defined]
        raise ValueError(f"{v} must be after Jan 1 2000")
    if pendulum.parse("3000-01-01T00:00:00Z").int_timestamp * 1000 < v:  # type: ignore[attr-defined]
        raise ValueError(f"{v} must be before Jan 1 3000")


class GtDispatchBoolean(BaseModel):
    """.

    Boolean dispatch command sent over the cloud broker from one GNode (FromGNodeAlias/ FromGNodeId) to another (ToGNodeAlias). AboutNodeAlias is the node getting dispatched. It may be a GNode, but it can also be a SpaceHeatNode.
    """

    AboutNodeName: str = Field(
        title="The Spaceheat Node getting dispatched",
    )
    ToGNodeAlias: str = Field(
        title="GNodeAlias of the SCADA",
        description=" [More info](https://gridworks.readthedocs.io/en/latest/scada.html).",
    )
    FromGNodeAlias: str = Field(
        title="GNodeAlias of AtomicTNode",
        description=" [More info](https://gridworks.readthedocs.io/en/latest/atomic-t-node.html).",
    )
    FromGNodeInstanceId: str = Field(
        title="GNodeInstance of the AtomicTNode",
    )
    RelayState: int = Field(
        title="0 or 1",
    )
    SendTimeUnixMs: int = Field(
        title="Time the AtomicTNode sends the dispatch, by its clock",
    )
    TypeName: Literal["gt.dispatch.boolean"] = "gt.dispatch.boolean"
    Version: str = "110"

    @validator("AboutNodeName")
    def _check_about_node_name(cls, v: str) -> str:
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(
                f"AboutNodeName failed LeftRightDot format validation: {e}"
            )
        return v

    @validator("ToGNodeAlias")
    def _check_to_g_node_alias(cls, v: str) -> str:
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(f"ToGNodeAlias failed LeftRightDot format validation: {e}")
        return v

    @validator("FromGNodeAlias")
    def _check_from_g_node_alias(cls, v: str) -> str:
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(
                f"FromGNodeAlias failed LeftRightDot format validation: {e}"
            )
        return v

    @validator("FromGNodeInstanceId")
    def _check_from_g_node_instance_id(cls, v: str) -> str:
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"FromGNodeInstanceId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @validator("SendTimeUnixMs")
    def _check_send_time_unix_ms(cls, v: int) -> int:
        try:
            check_is_reasonable_unix_time_ms(v)
        except ValueError as e:
            raise ValueError(
                f"SendTimeUnixMs failed ReasonableUnixTimeMs format validation: {e}"
            )
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))  # noqa


class GtDispatchBoolean_Maker:
    type_name = "gt.dispatch.boolean"
    version = "110"

    def __init__(
        self,
        about_node_name: str,
        to_g_node_alias: str,
        from_g_node_alias: str,
        from_g_node_instance_id: str,
        relay_state: int,
        send_time_unix_ms: int,
    ):
        self.tuple = GtDispatchBoolean(
            AboutNodeName=about_node_name,
            ToGNodeAlias=to_g_node_alias,
            FromGNodeAlias=from_g_node_alias,
            FromGNodeInstanceId=from_g_node_instance_id,
            RelayState=relay_state,
            SendTimeUnixMs=send_time_unix_ms,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: GtDispatchBoolean) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> GtDispatchBoolean:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> GtDispatchBoolean:
        d2 = dict(d)
        if "AboutNodeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing AboutNodeName")
        if "ToGNodeAlias" not in d2.keys():
            raise SchemaError(f"dict {d2} missing ToGNodeAlias")
        if "FromGNodeAlias" not in d2.keys():
            raise SchemaError(f"dict {d2} missing FromGNodeAlias")
        if "FromGNodeInstanceId" not in d2.keys():
            raise SchemaError(f"dict {d2} missing FromGNodeInstanceId")
        if "RelayState" not in d2.keys():
            raise SchemaError(f"dict {d2} missing RelayState")
        if "SendTimeUnixMs" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SendTimeUnixMs")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return GtDispatchBoolean(
            AboutNodeName=d2["AboutNodeName"],
            ToGNodeAlias=d2["ToGNodeAlias"],
            FromGNodeAlias=d2["FromGNodeAlias"],
            FromGNodeInstanceId=d2["FromGNodeInstanceId"],
            RelayState=d2["RelayState"],
            SendTimeUnixMs=d2["SendTimeUnixMs"],
            TypeName=d2["TypeName"],
            Version="110",
        )



# type: ignore
import time
from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

import pendulum
import pytz

from gwatn.data_classes import DEdge
from gwatn.data_classes import DNode
from gwatn.enums import RecognizedCurrencyUnit


class Edges:
    """A doubly indexed dict of DEDges"""

    as_dict: Dict[int, Dict[int, DEdge]]

    def __repr__(self) -> str:
        return (
            "calling as_dict returns a double dictionary of edges from a Node  \n"
            "with first key of the start_ts_idx of the Node and second key of start_idx"
        )


class DGraph(ABC):
    def __init__(
        self,
        d_graph_id: str,
        graph_strategy_alias: str,
        flo_start_unix_time_s: int,
        slice_duration_hrs: List[float],
        timezone_string: str,
        default_storage_steps: int,
        starting_store_idx: int,
        home_city: str,
        currency_unit: RecognizedCurrencyUnit,
        max_storage: float,
        max_power_in: float,
        wh_exponent: int = 3,
    ):
        self.graph_strategy_alias = graph_strategy_alias
        self.starting_store_idx = starting_store_idx
        self.edges: Dict[
            DNode, List[DEdge]
        ] = (
            {}
        )  # a dictionary that takes nodes to lists of edges connecting that node to nodes in the next time slice
        self.best_edge: Dict[
            DNode, DEdge
        ] = {}  # a dictionary that takes node to its best edge in the dijsktra path
        self.node: Dict[
            int, Dict[int, DNode]
        ] = (
            {}
        )  # self.node[jj][kk] is the node with time slice index jj and energy index kk
        self.slice_duration_hrs = slice_duration_hrs
        self.time_slices: int = len(self.slice_duration_hrs)
        self.default_storage_steps = default_storage_steps
        self.storage_steps: Dict[int, int] = {}
        self.set_storage_steps_per_time_slice()
        self.flo_start_unix_time_s: int = flo_start_unix_time_s
        self.timezone_string: str = timezone_string
        self.validate_timezone_format()
        self.flo_start_utc: Any = pendulum.from_timestamp(self.flo_start_unix_time_s)
        self.slice_start_utc: Dict[int, Any] = {}
        self.slice_start_unix_seconds: Dict[int, int] = {}
        self.set_slice_start()

        self.home_city = home_city
        self.currency_unit = currency_unit
        self.max_storage = max_storage
        self.max_power_in = max_power_in
        self.wh_exponent = wh_exponent
        self.d_graph_id = d_graph_id

    def __repr__(self) -> str:
        return f"Flo: Starting node: {self.node[0][self.starting_store_idx]}. Try flo.edges[flo.node[0][flo.starting_idx]"

    def validate_timezone_format(self) -> None:
        if not self.timezone_string in pytz.all_timezones:
            raise Exception(
                f"self.timezone_string '{self.timezone_string}' is not in the list of pytz.all_timezones"
            )

    @abstractmethod
    def create_graph(self) -> None:
        """The Dijsktra graph has nodes indexed first by
        time_slice_idx (range(self.time_slices)) and then by
        store_idx (range(self.store_steps[time_slice_idx])).
        The edges are always between nodes in adjacent time_slices,
        and reflect the potential future state changes that could
        occur. Each edge is, importantly, assigned an edge cost. In
        addition nodes can have `local benefits` (i.e. negative costs).

        Building this graph takes more time than solving for the
        lowest-cost path through the graph - which is Dijsktra's
        algorithm.

        Raises:
            NotImplementedError: This method is expected to
            be created by individual flos
        """
        raise NotImplementedError

    def set_final_slice_benefits(self) -> None:
        for node in self.node[self.time_slices].values():
            node.path_benefit = node.local_benefit
            node.path_edge_cost_only = 0

    def solve_dijkstra(self) -> None:
        st = time.time()
        # print("Solving dijkstra")
        self.set_final_slice_benefits()
        last_bad_slice_found = False
        for mm in range(1, self.time_slices + 1):
            # start from the last time slice and work to the front
            ts_idx = self.time_slices - mm
            bad_slice = True
            for node in self.node[ts_idx].values():
                edges = self.edges[node]
                # options gives a list of edge choices along with the cost of the path
                # from the node assuming that this edge is chosen, and that from that point
                # on the optimal path is chosen.

                # The path cost is the sum of the local cost at the current node, the
                # edge cost of the edge choice, and the path cost of the node in the next
                # time slice.
                try:
                    options = list(
                        map(
                            lambda x: [
                                x,
                                self.node[ts_idx + 1][x.end_idx].path_benefit
                                + self.node[ts_idx][x.start_idx].local_benefit
                                - x.cost,
                            ],
                            edges,
                        )
                    )
                except:
                    raise Exception(f"failed for node {node}")
                try:
                    best: Tuple[int, DEdge] = max(options, key=lambda x: x[1])
                except ValueError:
                    raise Exception(f"trouble with node {node}")

                best_edge: DEdge = best[0]
                self.best_edge[node]: DEdge = best_edge

                node.path_benefit = best[1]
                best_next_node = self.node[ts_idx + 1][best_edge.end_idx]
                node.best_next_idx = best_next_node.store_idx
                node.path_edge_cost_only = (
                    best_edge.cost + best_next_node.path_edge_cost_only
                )
                if node.path_cost < 10**5:
                    bad_slice = False

            if bad_slice is True and last_bad_slice_found is False:
                # print(f"last bad slice found: {ts_idx}")
                last_bad_slice_found = True
        # print(f"Total time for solving dijkstra:{1000*(time.time() - st):2.0f} ms ")

    def set_storage_steps_per_time_slice(self) -> None:
        """Initialize the dictionary self.store_steps.
        The keys are range(self.time_slices) and the values
        are the number of storage steps (e.g. how many
        ESteps) at the beginning of that time slice.

        This can be overwritten by inheriting classes
        """
        for jj in range(self.time_slices):
            self.storage_steps[jj] = self.default_storage_steps

    def set_slice_start(self) -> None:
        sd = self.flo_start_utc
        self.slice_start_utc[0] = sd
        self.slice_start_unix_seconds[0] = int(sd.timestamp())
        for i in range(1, self.time_slices):
            sd += pendulum.duration(hours=self.slice_duration_hrs[i - 1])
            self.slice_start_utc[i] = sd
            self.slice_start_unix_seconds[i] = int(sd.timestamp())

    @property
    def year_start_utc(self) -> int:
        year_start_in_local_time = pendulum.datetime(
            self.flo_start_utc.year, 1, 1, 0, 0, 0, tz=self.timezone_string
        )
        utc_year_start = pendulum.datetime(self.flo_start_utc.year, 1, 1, 0, 0, 0)
        utc_delta_hrs = (
            year_start_in_local_time - utc_year_start
        ).total_seconds() / 3600
        year_start = utc_year_start + pendulum.duration(hours=utc_delta_hrs)
        return year_start



""" GwSchema Enums used in gwatn """

# From gridworks
from gridworks.enums.algo_cert_type import AlgoCertType
from gridworks.enums.core_g_node_role import CoreGNodeRole
from gridworks.enums.g_node_role import GNodeRole
from gridworks.enums.g_node_status import GNodeStatus
from gridworks.enums.gni_status import GniStatus
from gridworks.enums.market_price_unit import MarketPriceUnit
from gridworks.enums.market_quantity_unit import MarketQuantityUnit
from gridworks.enums.market_type_name import MarketTypeName
from gridworks.enums.message_category import MessageCategory
from gridworks.enums.message_category_symbol import MessageCategorySymbol
from gridworks.enums.recognized_currency_unit import RecognizedCurrencyUnit
from gridworks.enums.strategy_name import StrategyName
from gridworks.enums.supervisor_container_status import SupervisorContainerStatus
from gridworks.enums.universe_type import UniverseType
from gwproto.enums.actor_class import ActorClass

# From gwproto
from gwproto.enums.local_comm_interface import LocalCommInterface
from gwproto.enums.make_model import MakeModel
from gwproto.enums.role import Role
from gwproto.enums.telemetry_name import TelemetryName
from gwproto.enums.unit import Unit

# From gwatn
from gwatn.enums.distribution_tariff import DistributionTariff
from gwatn.enums.energy_supply_type import EnergySupplyType
from gwatn.enums.recognized_temperature_unit import RecognizedTemperatureUnit


__all__ = [
    "AlgoCertType",
    "CoreGNodeRole",
    "DistributionTariff",
    "EnergySupplyType",
    "GNodeRole",
    "GNodeStatus",
    "GniStatus",
    "LocalCommInterface",
    "MarketPriceUnit",
    "MarketQuantityUnit",
    "MarketTypeName",
    "MessageCategory",
    "MessageCategorySymbol",
    "RecognizedCurrencyUnit",
    "RecognizedTemperatureUnit",
    "ActorClass",
    "Role",
    "MakeModel",
    "TelemetryName",
    "Unit",
    "StrategyName",
    "SupervisorContainerStatus",
    "UniverseType",
]



from typing import Optional

from pydantic import BaseModel

from gwatn.brick_storage_heater.flo import Flo__BrickStorageHeater as Flo
from gwatn.types import AtnBid
from gwatn.types import AtnParamsBrickstorageheater as AtnParams
from gwatn.types import FloParamsBrickstorageheater as FloParams
from gwatn.types import MarketSlot


class SlotStuff(BaseModel):
    Slot: MarketSlot
    BidParams: Optional[FloParams] = None
    Flo: Optional[Flo] = None
    Bid: Optional[AtnBid] = None
    Price: Optional[float] = None


def dummy_bid() -> AtnBid:
    return AtnBid(
        BidderAlias="d1.isone.dummy.ta",
    )


def dummy_slot_stuff(slot: MarketSlot) -> SlotStuff:
    return SlotStuff(
        Slot=slot, FloStartIdx=50, BidParams=dummy_flo_params(), AtnBid=dummy_bid()
    )


def is_dummy_slot_stuff(bid_stuff: SlotStuff) -> bool:
    if is_dummy_flo_params(bid_stuff.BidParams):
        return True
    return False


def is_dummy_atn_params(atn_params: AtnParams) -> bool:
    if atn_params.GNodeAlias == "d1.isone.dummy.ta":
        return True
    return False


def dummy_atn_params() -> AtnParams:
    return AtnParams(
        SliceDurationMinutes=60,
        FloSlices=48,
        GNodeAlias="d1.isone.dummy.ta",
        GNodeInstanceId="00000000-0000-0000-0000-000000000000",
        TypeName="atn.params.heatpumpwithbooststore",
        Version="000",
    )


def is_dummy_flo_params(flo_params: FloParams) -> bool:
    if flo_params.RtElecPriceUid == "00000000-0000-0000-0000-000000000000":
        return True
    return False


def dummy_flo_params() -> FloParams:
    return FloParams(
        RtElecPriceUid="00000000-0000-0000-0000-000000000000",
        WeatherUid="00000000-0000-0000-0000-000000000000",
    )


def get_max_store_kwh_th(params: FloParams) -> float:
    """Use max store temp, room temp, and CF to get max store energy in kWh_th"""
    room_temp_c = (params.RoomTempF - 32) * 5 / 9
    return (params.MaxBrickTempC - room_temp_c) * params.C


def get_house_worst_case_heat_output_avg_kw(params: FloParams) -> float:
    design_t = params.HouseWorstCaseTempF
    this_run_t = min(params.OutsideTempF)
    room_t = params.RoomTempF
    p = params.PowerRequiredByHouseFromSystemAvgKwList
    this_run_max_system_kw = max(p)
    this_run_max_kw_in = this_run_max_system_kw + params.AmbientPowerInKw
    dd_max_kw_in = this_run_max_kw_in * (room_t - design_t) / (room_t - this_run_t)
    return dd_max_kw_in - params.AmbientPowerInKw



""" Heatpumpwithbooststore Flo Edge Class Definition """
from typing import Optional
from typing import no_type_check

from gwatn.data_classes.d_edge import DEdge


SIG_FIGS_FOR_OUTPUT = 6


class Edge__BrickStorageHeater(DEdge):
    def __init__(
        self,
        start_ts_idx: int,
        start_idx: int,
        end_idx: int,
        avg_kw: Optional[float] = None,
        cost: Optional[float] = None,
    ):
        DEdge.__init__(
            self,
            start_ts_idx=start_ts_idx,
            start_idx=start_idx,
            end_idx=end_idx,
            cost=cost,
        )
        self.avg_kw = avg_kw

    def __repr__(self) -> str:
        rep = f"DEdge => Time Slice Idx: {self.start_ts_idx}, StartIdx: {self.start_idx}, EndIdx: {self.end_idx}"
        if self.cost is not None:
            rep += f" Cost => {self.cost}"
        return rep



"""Type initial.tadeed.algo.transfer, version 000"""
import json
from typing import Any
from typing import Dict
from typing import Literal

from gridworks.errors import SchemaError
from pydantic import BaseModel
from pydantic import Field
from pydantic import root_validator
from pydantic import validator


def check_is_algo_address_string_format(v: str) -> None:
    """
    AlgoAddressStringFormat format: The public key of a private/public Ed25519
    key pair, transformed into an  Algorand address, by adding a 4-byte checksum
    to the end of the public key and then encoding in base32.

    Raises:
        ValueError: if not AlgoAddressStringFormat format
    """
    import algosdk

    at = algosdk.abi.AddressType()
    try:
        result = at.decode(at.encode(v))
    except Exception as e:
        raise ValueError(f"Not AlgoAddressStringFormat: {e}")


def check_is_algo_msg_pack_encoded(v: str) -> None:
    """
    AlgoMSgPackEncoded format: the format of an  transaction sent to
    the Algorand blockchain.

    Raises:
        ValueError: if not AlgoMSgPackEncoded  format
    """
    import algosdk

    try:
        algosdk.encoding.future_msgpack_decode(v)
    except Exception as e:
        raise ValueError(f"Not AlgoMsgPackEncoded format: {e}")


class InitialTadeedAlgoTransfer(BaseModel):
    """TaValidator sends to GNodeFactory after validating Transactive Device.

    Once the TaValidator has done the initial on-site inspection of the Transactive
    Device including its location and the type and quality  of its power and energy
    metering, the TaValidator lets the GNodeFactory know by sending this message.
    Note the message also  includes the lat/lon of the Transactive Device. On
    receiving and validating this message, the GNodeFactory will co-sign the
    transfer and send the TaDeed to the TaDaemon address. In addition, the
    GNodeFactory creates and sends a TaTradingRights certificate to the
    TaDaemon address. Only once the GNodeFactory has verified that the
    TaDaemon address owns the TaDeed and TaTradingRights will it change
    the GNodeStatus of the associated TerminalAsset from  Pending to Active.
    [GNodeStatus](https://gridworks.readthedocs.io/en/latest/g-node-status.html)
    [TaDeed](https://gridworks.readthedocs.io/en/latest/ta-deed.html)
    [TaTradingRights](https://gridworks.readthedocs.io/en/latest/ta-trading-rights.html)
    [TaValidator](https://gridworks.readthedocs.io/en/latest/ta-validator.html)
    [TerminalAsset](https://gridworks.readthedocs.io/en/latest/terminal-asset.html)
    [Transactive Device](https://gridworks.readthedocs.io/en/latest/transactive-device.html)


    """

    MicroLat: int = Field(
        title="MicroLat",
        description="The Latitude of the Transactive Device, times 10^6",
    )
    MicroLon: int = Field(
        title="MicroLon",
        description="The Longitude of the Transactive Device, times 10^6",
    )
    ValidatorAddr: str = Field(
        title="ValidatorAddr",
        description="The Algoand address for the TaValidator who validated the location, metering and type of the Transactive Device. [More info](https://gridworks.readthedocs.io/en/latest/ta-validator.html).",
    )
    TaDaemonAddr: str = Field(
        title="TaDaemonAddr",
        description="The Algorand address for the TaDaemon which will own the TaDeed and initially the TaTradingRights), as well as holding funds on behalf of the TaOwner. [More info](https://gridworks.readthedocs.io/en/latest/ta-daemon.html).",
    )
    TaOwnerAddr: str = Field(
        title="TaOwnerAddr",
        description="The Algorand address of the entity owning the Transactive Device, and thus also the TerminalAsset",
    )
    FirstDeedTransferMtx: str = Field(
        title="FirstDeedTransferMtx",
        description="The half-signed multi transaction for transferring the TaDeed to the TaDaemon.",
    )
    TypeName: Literal["initial.tadeed.algo.transfer"] = "initial.tadeed.algo.transfer"
    Version: str = "000"

    @validator("ValidatorAddr")
    def _check_validator_addr(cls, v: str) -> str:
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"ValidatorAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("TaDaemonAddr")
    def _check_ta_daemon_addr(cls, v: str) -> str:
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"TaDaemonAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("TaOwnerAddr")
    def _check_ta_owner_addr(cls, v: str) -> str:
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"TaOwnerAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("FirstDeedTransferMtx")
    def _check_first_deed_transfer_mtx(cls, v: str) -> str:
        try:
            check_is_algo_msg_pack_encoded(v)
        except ValueError as e:
            raise ValueError(
                f"FirstDeedTransferMtx failed AlgoMsgPackEncoded format validation: {e}"
            )
        return v

    @root_validator(pre=True)
    def check_axiom_1(cls, v: dict) -> dict:
        """
        Axiom 1: Is correct Multisig.
        Decoded FirstDeedTransferMtx must have type MultisigTransaction from the
        2-sig MultiAccount  [GnfAdminAddr, ValidatorAddr].
        [More info](https://gridworks.readthedocs.io/en/latest/g-node-factory.html#gnfadminaddr)
        """
        # TODO: Implement check for axiom 1"
        return v

    @root_validator
    def check_axiom_2(cls, v: dict) -> dict:
        """
        Axiom 2: TaDaemon funded by TaOwner.
        The TaDaemonAddr was created with funding from the TaOwnerAddr, and has sufficient funding according to the GNodeFactory.
        """
        # TODO: Implement check for axiom 2"
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))  # noqa


class InitialTadeedAlgoTransfer_Maker:
    type_name = "initial.tadeed.algo.transfer"
    version = "000"

    def __init__(
        self,
        micro_lat: int,
        micro_lon: int,
        validator_addr: str,
        ta_daemon_addr: str,
        ta_owner_addr: str,
        first_deed_transfer_mtx: str,
    ):
        self.tuple = InitialTadeedAlgoTransfer(
            MicroLat=micro_lat,
            MicroLon=micro_lon,
            ValidatorAddr=validator_addr,
            TaDaemonAddr=ta_daemon_addr,
            TaOwnerAddr=ta_owner_addr,
            FirstDeedTransferMtx=first_deed_transfer_mtx,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: InitialTadeedAlgoTransfer) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> InitialTadeedAlgoTransfer:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> InitialTadeedAlgoTransfer:
        d2 = dict(d)
        if "MicroLat" not in d2.keys():
            raise SchemaError(f"dict {d2} missing MicroLat")
        if "MicroLon" not in d2.keys():
            raise SchemaError(f"dict {d2} missing MicroLon")
        if "ValidatorAddr" not in d2.keys():
            raise SchemaError(f"dict {d2} missing ValidatorAddr")
        if "TaDaemonAddr" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TaDaemonAddr")
        if "TaOwnerAddr" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TaOwnerAddr")
        if "FirstDeedTransferMtx" not in d2.keys():
            raise SchemaError(f"dict {d2} missing FirstDeedTransferMtx")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return InitialTadeedAlgoTransfer(
            MicroLat=d2["MicroLat"],
            MicroLon=d2["MicroLon"],
            ValidatorAddr=d2["ValidatorAddr"],
            TaDaemonAddr=d2["TaDaemonAddr"],
            TaOwnerAddr=d2["TaOwnerAddr"],
            FirstDeedTransferMtx=d2["FirstDeedTransferMtx"],
            TypeName=d2["TypeName"],
            Version="000",
        )



from gwproto import Decoders
from gwproto import MQTTCodec
from gwproto import create_message_payload_discriminator
from gwproto.messages import GtShStatus_Maker
from gwproto.messages import PowerWatts_Maker
from gwproto.messages import SnapshotSpaceheat_Maker


ScadaMessageDecoder = create_message_payload_discriminator(
    model_name="GwprotoMessageDecoder", module_names=["gwproto.messages"]
)


class ScadaCodec(MQTTCodec):
    def __init__(self):
        super().__init__(
            Decoders.from_objects(
                [
                    GtShStatus_Maker,
                    SnapshotSpaceheat_Maker,
                    PowerWatts_Maker,
                ],
                message_payload_discriminator=ScadaMessageDecoder,
            )
        )

    def validate_source_alias(self, source_alias: str):
        pass



import csv
import json
import uuid

import pendulum

from gwatn.types import AtnParamsBrickstorageheater as AtnParams
from gwatn.types import AtnParamsBrickstorageheater_Maker as AtnParams_Maker
from gwatn.types import AtnParamsReport_Maker
from gwatn.types.csv_distp_sync import CsvDistpSync_Maker
from gwatn.types.csv_eprt_sync import CsvEprtSync_Maker
from gwatn.types.csv_weather_forecast_sync import CsvWeatherForecastSync_Maker


eprt_csv = "input_data/elec_price_data.csv"
ep = CsvEprtSync_Maker(elec_price_file=eprt_csv).tuple

distp_csv = "input_data/dist_price_data.csv"
dp = CsvDistpSync_Maker(dist_price_file=distp_csv).tuple

temp_csv = "input_data/temperature_data.csv"
temp = CsvWeatherForecastSync_Maker(weather_csv=temp_csv).tuple

params_file = "src/gwatn/brick_storage_heater/dev_atn_params_data.csv"
params = []
with open(params_file) as csv_file:
    reader = csv.reader(csv_file)
    for row in reader:
        params.append(row)
start = pendulum.datetime(year=2020, month=1, day=1, hour=4)

for j in range(1, 19):
    alias = params[1][j]
    atn_params = AtnParams(
        GNodeAlias=alias,
        MaxBrickTempC=params[3][j],
        RatedMaxPowerKw=params[4][j],
        C=params[5][j],
        ROff=params[6][j],
        ROn=params[7][j],
        RoomTempF=params[8][j],
        AnnualHvacKwhTh=params[9][j],
        BetaOt=params[10][j],
    )

    try:
        report = AtnParamsReport_Maker(
            g_node_alias=alias,
            g_node_instance_id=str(uuid.uuid4()),
            atn_params_type_name=AtnParams_Maker.type_name,
            time_unix_s=start.int_timestamp,
            params=atn_params,
            irl_time_unix_s=None,
        ).tuple
    except:
        print(f"Problem with {alias}")

    file_name = f"input_data/eventstore/{alias}-{AtnParamsReport_Maker.type_name}-{start.int_timestamp * 1000}.json"
    json_object = json.dumps(report.as_dict())

    # Writing to sample.json
    with open(file_name, "w") as outfile:
        outfile.write(json_object)
